# _bibfiles.py - ordered collection of bibfiles with load/save api

import os
import io
import datetime

from clldutils.inifile import INI

from pyglottolog.util import references_path
import _bibtex
from _bibfiles_db import Database

__all__ = ['Collection', 'BibFile', 'Database']

DIR = references_path('bibtex').as_posix()
CONFIG = 'BIBFILES.ini'


class Collection(list):
    """Directory with an INI-file with settings for BibTeX files inside."""

    _encoding = 'utf-8-sig'

    @classmethod
    def _bibfiles(cls, directory, config, endwith):
        """Read the INI-file, yield bibfile instances for sections."""
        cfg = INI(interpolation=None)
        cfg.read(os.path.join(directory, '..', config))
        for s in cfg.sections():
            if not s.endswith(endwith):
                continue
            filepath = os.path.join(directory, s)
            assert os.path.exists(filepath)
            sortkey = cfg.get(s, 'sortkey')
            if sortkey.lower() == 'none':
                sortkey = None
            yield BibFile(
                filepath=filepath,
                encoding=cfg.get(s, 'encoding'), sortkey=sortkey,
                use_pybtex=cfg.getboolean(s, 'use_pybtex'),
                priority=cfg.getint(s, 'priority'),
                name=cfg.get(s, 'name'), title=cfg.get(s, 'title'),
                description=cfg.get(s, 'description'),
                abbr=cfg.get(s, 'abbr'))

    def __init__(self, directory=DIR, config=CONFIG, endwith='.bib'):
        self.directory = directory
        bibfiles = self._bibfiles(directory, config, endwith)
        super(Collection, self).__init__(bibfiles)
        self._map = {b.filename: b for b in self}

    def __getitem__(self, index_or_filename):
        """Retrieve a bibfile by index or filename."""
        if isinstance(index_or_filename, basestring):
            return self._map[index_or_filename]
        return super(Collection, self).__getitem__(index_or_filename)

    def to_sqlite(self, filename=None, rebuild=False):
        """Return a database with the bibfiles loaded."""
        return Database.from_bibfiles(self, filename, rebuild=rebuild)

    def check_all(self):
        """Check the BibTeX syntax of all bibfiles."""
        for b in self:
            b.check()
            
    def roundtrip_all(self):
        """Load and save all bibfiles with the current settings."""
        for b in self:
            b.roundtrip()


class BibFile(object):
    """BibTeX source file with configurable load/save options and meta data."""

    def __init__(self, filepath, encoding, sortkey, use_pybtex=True, priority=0,
                 name=None, title=None, description=None, abbr=None):
        self.filepath = filepath
        self.filename = os.path.basename(filepath)
        self.encoding = encoding
        self.sortkey = sortkey
        self.use_pybtex = use_pybtex
        self.priority = priority
        self.name = name
        self.title = title
        self.description = description
        self.abbr = abbr

    @property
    def size(self):
        return os.stat(self.filepath).st_size

    @property
    def mtime(self):
        return datetime.datetime.fromtimestamp(os.stat(self.filepath).st_mtime)

    def iterentries(self):
        """Yield entries as (bibkey, (entrytype, fields)) tuples."""
        return _bibtex.iterentries(filename=self.filepath,
            encoding=self.encoding,
            use_pybtex=self.use_pybtex)

    def load(self):
        """Return entries as bibkey -> (entrytype, fields) dict."""
        return _bibtex.load(filename=self.filepath,
            preserve_order=self.sortkey is None,
            encoding=self.encoding,
            use_pybtex=self.use_pybtex)

    def save(self, entries, verbose=True):
        """Write bibkey -> (entrytype, fields) map to file."""
        _bibtex.save(entries,
            filename=self.filepath,
            sortkey=self.sortkey,
            encoding=self.encoding,
            use_pybtex=self.use_pybtex,
            verbose=verbose)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.filename)

    def check(self):
        print(self)
        entries = self.load()  # bare BibTeX syntax
        invalid = _bibtex.check(filename=self.filepath)  # names/macros etc.
        verdict = ('(%d invalid)' % invalid) if invalid else 'OK'
        print('%d %s' % (len(entries), verdict))

    def roundtrip(self):
        print(self)
        self.save(self.load())

    def show_characters(self, include_plain=False):
        """Display character-frequencies (excluding printable ASCII)."""
        import collections
        from unicodedata import name

        with io.open(self.filepath, encoding=self.encoding) as fd:
            data = fd.read()
        hist = collections.Counter(data)
        table = '\n'.join('%d\t%-9r\t%s\t%s' % (n, c, c, name(c, ''))
            for c, n in hist.most_common()
            if include_plain or not 20 <= ord(c) <= 126)
        print(table)


if __name__ == '__main__':
    c = Collection()
    d = Database()



# coding: utf8
from __future__ import unicode_literals
from unittest import TestCase
import re


class Tests(TestCase):
    def test_ID_REGEX(self):
        from pyglottolog.languoids import ID_REGEX

        pattern = re.compile(ID_REGEX + '$')
        for valid in [
            'abcd1234',
            'a12d3456',
            'NOCODE',
            'NOCODE_abd',
            'NOCODE_1',
        ]:
            self.assertIsNotNone(pattern.match(valid))

        for invalid in [
            'abcd123',
            '12d3456',
            'aNOCODE',
            'NOCODE.abd',
            'nocode',
        ]:
            self.assertIsNone(pattern.match(invalid))


# bibtex.py - bibtex file parsing/serialization

# TODO: make check fail on non-whitespace between entries (bibtex 'comments')

import io
import re
import mmap
import contextlib
import collections

from pybtex.database.input.bibtex import BibTeXEntryIterator, Parser, UndefinedMacro
from pybtex.scanner import PybtexSyntaxError
from pybtex.exceptions import PybtexError
from pybtex.textutils import whitespace_re
from pybtex.bibtex.utils import split_name_list
from pybtex.database import Person

from _bibtex_escaping import u_escape, u_unescape, latex_to_utf8

__all__ = [
    'load', 'iterentries', 'names',
    'save', 'dump',
    'check',
]

FIELDORDER = [
    'author', 'editor', 'title', 'booktitle', 'journal',
    'school', 'publisher', 'address',
    'series', 'volume', 'number', 'pages', 'year', 'issn', 'url',
]

VERBATIM = {'doi', 'eprint', 'file', 'url', 'pdf', 'fn', 'fnnote'}


@contextlib.contextmanager
def memorymapped(filename, access=mmap.ACCESS_READ):
    fd = open(filename)
    try:
        m = mmap.mmap(fd.fileno(), 0,  access=access)
    except:
        fd.close()
        raise
    try:
        yield m
    finally:
        m.close()
        fd.close()


def load(filename, preserve_order=False, encoding=None, use_pybtex=True):
    cls = collections.OrderedDict if preserve_order else dict
    return cls(iterentries(filename, encoding, use_pybtex))


def iterentries(filename, encoding=None, use_pybtex=True):
    if not use_pybtex:  # legacy code path for conversion/comparison
        if encoding not in (None, 'ascii'):
            raise NotImplementedError
        import _libmonster
        with memorymapped(filename) as source:
            for bibkey, entrytype, fields in _libmonster.pitems(source):
                yield bibkey, (entrytype, fields)  
    elif encoding is None:
        raise NotImplementedError
    else:
        with memorymapped(filename) as source:
            try:
                for entrytype, (bibkey, fields) in BibTeXEntryIterator(source):
                    fields = {name.decode(encoding).lower():
                        whitespace_re.sub(' ', ''.join(values).decode(encoding).strip())
                        for name, values in fields}
                    yield bibkey.decode(encoding), (entrytype.decode(encoding), fields)
            except PybtexSyntaxError as e:
                debug_pybtex(source, e)


def debug_pybtex(source, e):
    start, line, pos = e.error_context_info
    print('BIBTEX ERROR on line %d, last parsed lines:' % line)
    print(source[start:start+500] + '...')
    raise
    

def names(s):
    for name in split_name_list(s):
        try:
            yield Name.from_string(name)
        except PybtexError as e:
            print(repr(e))


class Name(collections.namedtuple('Name', 'prelast last given lineage')):

    __slots__ = ()

    @classmethod
    def from_string(cls, name):
        person = Person(name)
        prelast, last, first, middle, lineage = (' '.join(getattr(person, part))
            for part in ('_prelast', '_last', '_first', '_middle', '_lineage'))
        given = ' '.join(n for n in (first, middle) if n)
        return cls(prelast, last, given, lineage)


def save(entries, filename, sortkey, encoding=None, errors='strict', use_pybtex=True, verbose=True):
    if encoding in (None, 'ascii', 'ascii+u_escape'):
        with open(filename, 'w') as fd:
            dump(entries, fd, sortkey, encoding, errors, use_pybtex, verbose)
    else:
        assert errors == 'strict'
        with io.open(filename, 'w', encoding=encoding, errors=errors) as fd:
            dump(entries, fd, sortkey, encoding, None, use_pybtex, verbose)


def dump(entries, fd, sortkey=None, encoding=None, errors='strict', use_pybtex=True, verbose=True, verbatim=VERBATIM):
    if sortkey is None:
        if isinstance(entries, collections.OrderedDict):
            items = entries.iteritems()
        elif isinstance(entries, dict):
            raise ValueError('dump needs sortkey or ordered entries')
        else:
            items = entries
    elif sortkey == 'bibkey':
        items = ((bibkey, entries[bibkey]) for bibkey in
            sorted(entries, key=lambda bibkey: bibkey.lower()))
    elif sortkey == 'authorbibkey_colon':  # legacy order for hh.bib
        def sortkey((bibkey, (entrytype, fields))):
            return fields.get('author', '') + ':'.join(bibkey.split(':', 1)[::-1])
        items = sorted(entries.iteritems(), key=sortkey)
    else:
        raise ValueError(sortkey)
    """Reserved characters (* -> en-/decoded by latexcodec)
    * #: \#
      $: \$
      %: \%
      ^: \^{} \textasciicircum
    * &: \&
    * _: \_
      {: \{
      }: \}
      ~: \~{} \textasciitilde
      \: \textbackslash{}
      <: \textless
      >: \textgreater
    """
    if not use_pybtex:  # legacy code path for conversion/comparison
        if encoding not in (None, 'ascii'):
            raise NotImplementedError
        for bibkey, (entrytype, fields) in items:
            fd.write('@%s{%s' % (entrytype, bibkey))
            for k, v in fieldorder.itersorted(fields):
                if k in verbatim:
                    v = v.strip().encode('ascii', errors)
                else:
                    v = v.strip().encode('latex', errors).replace(r'\#', '#').replace(r'\&', r'&').replace(r'\_', '_')
                fd.write(',\n    %s = {%s}' % (k, v))
            fd.write('\n}\n' if fields else ',\n}\n')
    elif encoding is None:
        raise NotImplementedError
    elif encoding == 'ascii+u_escape':
        for bibkey, (entrytype, fields) in items:
            fd.write('@%s{%s' % (entrytype, bibkey))
            for k, v in fieldorder.itersorted(fields):
                if k in verbatim:
                    v = v.strip().encode('ascii')
                else:
                    v = u_escape(v).strip().encode('latex', errors).replace(r'\#', '#').replace(r'\\&', r'\&').replace(r'\_', '_')
                fd.write(',\n    %s = {%s}' % (k, v))
            fd.write('\n}\n' if fields else ',\n}\n')
    elif encoding == 'ascii':
        for bibkey, (entrytype, fields) in items:
            fd.write('@%s{%s' % (entrytype, bibkey))
            for k, v in fieldorder.itersorted(fields):
                if k in verbatim:
                    v = v.strip().encode('ascii')
                else:
                    v = v.strip().encode('latex', errors).replace(r'\#', '#').replace(r'\\&', r'\&').replace(r'\_', '_')
                fd.write(',\n    %s = {%s}' % (k, v))
            fd.write('\n}\n' if fields else ',\n}\n')
    else:
        assert errors is None
        for bibkey, (entrytype, fields) in items:
            fd.write(u'@%s{%s' % (entrytype, bibkey))
            for k, v in fieldorder.itersorted(fields):
                if k in verbatim:
                    v = v.strip().decode('ascii')
                elif isinstance(v, str):
                    v = latex_to_utf8(v.strip(), verbose=verbose)
                fd.write(u',\n    %s = {%s}' % (k, v))
            fd.write(u'\n}\n' if fields else u',\n}\n')


class Ordering(dict):
    """Key order for iterating over dicts (unknown keys last alphabetic)."""

    _missing = float('inf')

    @classmethod
    def fromlist(cls, keys):
        """Define the order of keys as given."""
        return cls((k, i) for i, k in enumerate(keys))

    def itersorted(self, dct):
        """Iterate over dct (key, value) pairs in the defined order."""
        for key in sorted(dct, key=self._itersorted_key):
            yield key, dct[key]

    def _itersorted_key(self, key):
         return self[key], key

    def __missing__(self, key):
        return self._missing


fieldorder = Ordering.fromlist(FIELDORDER)


def check(filename, encoding=None):
    parser = CheckParser(encoding=encoding)
    parser.parse_file(filename)
    return parser.error_count
    

class CheckParser(Parser):
    """Unline BibTeXEntryIterator also parses names, macros, etc."""

    def __init__(self, *args, **kwargs):
        super(CheckParser, self).__init__(*args, **kwargs)
        self.error_count = 0

    def handle_error(self, error):
        print('%r' % error)
        self.error_count += 1
        if not isinstance(error, UndefinedMacro):
            raise error

    def process_entry(self, *args, **kwargs):
        try:
            super(CheckParser, self).process_entry(*args, **kwargs)
        except PybtexError as e:
            print(e)
            self.error_count += 1


def _test_load():
    import _bibfiles
    _bibfiles.Collection().check_all()


if __name__ == '__main__':
    _test_load()



# coding: utf8
from __future__ import unicode_literals
from unittest import TestCase
from contextlib import contextmanager

from pyglottolog.lff import read_lff


@contextmanager
def lff(text):
    yield text.split('\n')


class Tests(TestCase):
    def test_read_lff(self):
        _lff = lff("""
Name [abcd1234], Name2 [abcd1235]
   Lang [abcd1236][abc]
""")
        res = list(read_lff('language', fp=_lff))
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].name, 'Lang')
        self.assertEqual(len(res[0].lineage), 2)
        self.assertEqual(res[0].lineage[0], ('Name', 'abcd1234', 'family'))

    def test_read_lff_error(self):
        _lff = lff("""
Name [acd1234], Name2 [abcd1235]
   Lang [abcd1236][abc]
""")
        with self.assertRaises(ValueError):
            list(read_lff('language', fp=_lff))



# monster.py - load, combine, annotate, and save bibfiles collection

"""Compiling the monster.

This script read the INI-file in the references/bibtex directory and combines
all .bib files configured there into a file called monster-utf8.bib with some
deduplication and annotation in the process.

1.    The bibfiles are combined in the following manner
1.1   All bibfiles are parsed and loaded into an sqlite3 database
1.2   A hash (keyid) is computed for each bib-entry
1.3.  For each hash, any bib-entries with that hash are merged
1.3.1 The merging takes place such that the priority setting of the file in the
      INI-file (in case of ties, filename and bibkey) decides which value takes
      precedence (hh.bib has the highest priority).
      Some fields (like note) are the concatenation of all original fields.
      The merged entries link back to the original(s) in the added
      src/srctrickle field.

2.    A glottolog_ref_id needs to be assigned to each merged entry
2.1   If a group of entries with the same glottolog_ref_id (last run) is split
      up into different hashes (this run), the new hash group which is most
      similar to the old group retains the glottolog_ref_id
2.2   If entries with different glottolog_ref_ids (last run) are merged into
      a single hash (this run), the old ref_id groups which are not most similar
      to the new group will be marked as being replaced by the most similar one
2.3   Once splitting/merging is resolved, remaining entries get glottolog_ref_ids
2.3.1 New glottolog_ref_id:s (from the private area above 300000) are doled out
      to bib-entries which do not have one

3.    Four steps of annotation are added to the merged entries, but only if
      there isn't already such annotation
3.1   macro_area is added based on the lgcode field if any. The mapping between
      lgcode:s and macro_area:s are taken from "../languoids/lginfo.csv"
3.2   hhtype is added based on a small set of trigger words that may occur in
      the titles of bibentries which are taken from 'alt4hhtype.ini'. A hhtype
      is not inferred if it would change the "descriptive status" of a language
      taken from hh.bib.
3.3   lgcode is added based on a large and dirty set of trigger words that
      may/may not occur in the titles of bibentries which are taken from
      'alt4lgcode.ini'. A lgcode is not inferred if it would change the
      "descriptive status" of a language taken from hh.bib.
3.4   inlg is added based on a small set of trigger words that may occur in the
      titles of bibentries which are specified in "../references/alt4inlg.ini".

4.    The assigned glottolog_ref_id are burned back into the original bib:s

5.    A final monster-utf8.bib is written
"""

import time

from pyglottolog.util import build_path, references_path
from pyglottolog import _bibfiles
from pyglottolog import _libmonster as bib
from pyglottolog import languoids

BIBFILES = _bibfiles.Collection()
PREVIOUS = references_path('monster.csv').as_posix()
REPLACEMENTS = build_path('monster-replacements.json').as_posix()
MONSTER = _bibfiles.BibFile(
    build_path('monster-utf8.bib').as_posix(), encoding='utf-8', sortkey='bibkey')

HHTYPE = references_path('alt4hhtype.ini').as_posix()
MARKHHTYPE = build_path('monstermark-hht.txt').as_posix()
MARKLGCODE = build_path('monstermark-lgc.txt').as_posix()


def intersectall(xs):
    a = set(xs[0])
    for x in xs[1:]:
        a.intersection_update(x)
    return a


def markconservative(m, trigs, ref, outfn="monstermarkrep.txt", blamefield="hhtype"):
    mafter = markall(m, trigs)
    ls = bib.lstat(ref)
    #print bib.fd(ls.values())
    lsafter = bib.lstat_witness(mafter)
    log = []
    for (lg, (stat, wits)) in lsafter.iteritems():
        if not ls.get(lg):
            print lg, "lacks status", [mafter[k][1]['srctrickle'] for k in wits]
            continue
        if bib.hhtype_to_n[stat] > bib.hhtype_to_n.get(ls[lg]):
            log = log + [(lg, [(mafter[k][1].get(blamefield, "No %s" % blamefield), k, mafter[k][1].get('title', 'no title'), mafter[k][1]['srctrickle']) for k in wits], ls[lg])]
            for k in wits:
                (t, f) = mafter[k]
                if f.has_key(blamefield):
                    del f[blamefield]
                mafter[k] = (t, f)
    bib.write_csv_rows(((lg, was) + mis for (lg, miss, was) in log for mis in miss), outfn, dialect='excel-tab')
    return mafter


def markall(e, trigs, labelab=lambda x: x):
    clss = set(cls for (cls, _) in trigs.iterkeys())
    ei = dict((k, (typ, fields)) for (k, (typ, fields)) in e.iteritems() if [c for c in clss if not fields.has_key(c)])

    wk = {}
    for (k, (typ, fields)) in ei.iteritems():
        for w in bib.wrds(fields.get('title', '')):
            bib.setd(wk, w, k)

    u = {}
    it = bib.indextrigs(trigs)
    for (dj, clslabs) in it.iteritems():
        mkst = [wk.get(w, {}).iterkeys() for (stat, w) in dj if stat]
        mksf = [set(ei.iterkeys()).difference(wk.get(w, [])) for (stat, w) in dj if not stat]
        mks = intersectall(mkst + mksf)
        for k in mks:
            for cl in clslabs:
                bib.setd3(u, k, cl, dj)

    for (k, cd) in u.iteritems():
        (t, f) = e[k]
        f2 = dict((a, b) for (a, b) in f.iteritems())
        for ((cls, lab), ms) in cd.iteritems():
            a = ';'.join(' and '.join(('' if stat else 'not ') + w for (stat, w) in m) for m in ms)
            f2[cls] = labelab(lab) + ' (computerized assignment from "' + a + '")'
            e[k] = (t, f2)
    print "trigs", len(trigs)
    print "trigger-disjuncts", len(it)
    print "label classes", len(clss)
    print "unlabeled refs", len(ei)
    print "updates", len(u)
    return e


def macro_area_from_lgcode(m):
    lgd = languoids.macro_area_from_hid()

    def inject_macro_area((typ, fields)):
        mas = set(lgd[x] for x in bib.lgcode((typ, fields)) if x in lgd and lgd[x])
        if mas:
            fields['macro_area'] = ', '.join(sorted(mas))
        return (typ, fields)
    
    return dict((k, inject_macro_area(tf)) for k, tf in m.iteritems())


def main(bibfiles=BIBFILES, previous=PREVIOUS, replacements=REPLACEMENTS, monster=MONSTER):
    print '%s open/rebuild bibfiles db' % time.ctime()
    db = bibfiles.to_sqlite()

    print '%s compile_monster' % time.ctime()
    m = dict(db.merged())

    print '%s load hh.bib' % time.ctime()
    hhbib = bibfiles['hh.bib'].load()

    # Annotate with macro_area from lgcode when lgcode is assigned manually
    print '%s macro_area_from_lgcode' % time.ctime()
    m = macro_area_from_lgcode(m)

    # Annotate with hhtype
    print '%s annotate hhtype' % time.ctime()
    hht = dict(((cls, bib.expl_to_hhtype[lab]), v) for ((cls, lab), v) in bib.load_triggers(HHTYPE).iteritems())
    m = markconservative(m, hht, hhbib, outfn=MARKHHTYPE, blamefield="hhtype")

    # Annotate with lgcode
    print '%s annotate lgcode' % time.ctime()
    lgc = languoids.load_triggers()
    m = markconservative(m, lgc, hhbib, outfn=MARKLGCODE, blamefield="hhtype")

    # Annotate with inlg
    print '%s add_inlg_e' % time.ctime()
    m = bib.add_inlg_e(m)

    # Print some statistics
    print time.ctime()
    print "# entries", len(m)
    print "with lgcode", sum(1 for t, f in m.itervalues() if 'lgcode' in f)
    print "with hhtype", sum(1 for t, f in m.itervalues() if 'hhtype' in f)
    print "with macro_area", sum(1 for t, f in m.itervalues() if 'macro_area' in f)

    # Update the CSV with the previous mappings for later reference
    print '%s update_previous' % time.ctime()
    db.to_csvfile(previous)

    print '%s save_replacements' % time.ctime()
    db.to_replacements(replacements)

    # Trickling back
    print '%s trickle' % time.ctime()
    db.trickle()

    # Save
    print '%s save as utf8' % time.ctime()
    monster.save(m, verbose=False)

    print '%s done.' % time.ctime()


if __name__ == '__main__':
    main()



# _compare_bibtex - compare regex-based with pybtex bibfile parsing

import _bibfiles, _bibtex

for b in _bibfiles.Collection():
    print(b.filepath)
    d = _bibtex.load(b.filepath, encoding=None, use_pybtex=False)
    e = _bibtex.load(b.filepath, encoding='utf-8', use_pybtex=True)
    if d.keys() != e.keys():
        print sorted(set(d).symmetric_difference(e))
    for k in e:
        x = d[k][1]
        y = e[k][1]
        if x.keys() != y.keys():
            print k
            print repr(x.keys())
            print repr(y.keys())
        for field in x:
            if x[field].decode('utf-8') != y[field]:
                print k
                print field
                print repr(x[field])
                print repr(y[field])



from __future__ import unicode_literals
import re

from clldutils.path import walk, Path

from pyglottolog import util
from pyglottolog import languoids


ISO_CODE_PATTERN = re.compile('[a-z]{3}$')


class Glottolog(object):
    def __init__(self, repos=None):
        self.repos = Path(repos) if repos else util.DATA_DIR
        self.tree = util.languoids_path('tree', data_dir=self.repos)

    def languoid(self, id_):
        if ISO_CODE_PATTERN.match(id_):
            for l in languoids.walk_tree(tree=self.tree):
                if l.iso_code == id_:
                    return l
        else:
            for d in walk(self.tree, mode='dirs'):
                if d.name == id_:
                    return languoids.Languoid.from_dir(d)

    def languoids(self, ids=None):
        for d in walk(self.tree, mode='dirs'):
            if ids is None or d.name in ids:
                yield languoids.Languoid.from_dir(d)



# coding: utf8
from __future__ import unicode_literals
from functools import partial

from clldutils.path import Path
from clldutils.inifile import INI

import pyglottolog


DATA_DIR = Path(pyglottolog.__file__).parent.parent


def subdir_path(subdir, *comps, **kw):
    data_dir = kw.pop('data_dir', DATA_DIR)
    return data_dir.joinpath(subdir, *comps)


languoids_path = partial(subdir_path, 'languoids')
references_path = partial(subdir_path, 'references')
build_path = partial(subdir_path, 'build')


def read_ini(filename, interpolation=None):
    if isinstance(filename, Path):
        filename = filename.as_posix()
    p = INI(interpolation=interpolation)
    p.read(filename)
    return p


def parse_conjunctions(expr):
    return [
        (False, w[4:].strip()) if w.startswith('NOT ') else (True, w.strip())
        for w in expr.split(' AND ')]



from __future__ import unicode_literals, print_function, division
from collections import defaultdict
import re
import os

from clldutils.path import Path, copytree, as_posix

from pyglottolog.util import build_path
from pyglottolog.languoids import Languoid, walk_tree, TREE, ID_REGEX


NAME_AND_ID_REGEX = '([^\[]+)(\[' + ID_REGEX + '\])'


def rmtree(d, **kw):
    d = as_posix(d)
    for path in (os.path.join(d, f) for f in os.listdir(d)):
        if os.path.isdir(path):
            rmtree(path)
        else:
            os.unlink(path)
    os.rmdir(d)


def read_lff(level, fp=None):
    lang_line = re.compile('\s+' + NAME_AND_ID_REGEX + '(\[([a-z]{3})?\])$')
    class_line = re.compile(NAME_AND_ID_REGEX + '(,\s*' + NAME_AND_ID_REGEX + ')*$')
    isolate_line = re.compile('([^\[]+)(\[-isolate-\])$')

    path = None
    with fp or build_path('%sff.txt' % level[0]).open(encoding='utf8') as fp:
        for line in fp:
            line = line.rstrip()
            if line.startswith('#') or not line.strip():
                # ignore comments or empty lines
                continue
            match = lang_line.match(line)
            if match:
                assert path
                yield Languoid.from_lff(
                    None if path == 'isolate' else path, line.strip(), level)
            else:
                match = isolate_line.match(line)
                if match:
                    path = 'isolate'
                else:
                    # assert it matches a classification line!
                    if not class_line.match(line):
                        raise ValueError(line)
                    path = line.strip()


def lang2tree(lang, lineage, out, old_tree):
    groupdir = out

    for name, id_, level in lineage:
        groupdir = groupdir.joinpath(id_)
        if not groupdir.exists():
            groupdir.mkdir()
            if id_ in old_tree:
                group = old_tree[id_]
                assert group.level == level
                if name != group.name:
                    # rename a subgroup!
                    group.name = name
            else:
                assert id_.startswith('NOCODE')
                group = Languoid.from_name_id_level(name, id_, level)
            group.write_info(groupdir)

    langdir = groupdir.joinpath(lang.fname())
    langdir.mkdir()

    if lang.id in old_tree:
        old_lang = old_tree[lang.id]
        assert old_lang.level == lang.level
        if old_lang.name != lang.name:
            old_lang.name = lang.name
        old_lang.write_info(langdir)
    else:
        assert lang.id.startswith('NOCODE')
        lang.write_info(langdir)


def lff2tree(tree=TREE, outdir=None, test=False):
    """
    - get mapping glottocode -> Languoid from old tree
    - assemble new directory tree
      - for each path component in lff/dff:
        - create new dir
        - copy info file from old tree (possibly updating the name) or
        - create info file
      - for each language/dialect in lff/dff:
        - create new dir
        - copy info file from old tree (possibly updating the name) or
        - create info file
    - rm old tree
    - copy new tree
    """
    out = Path(outdir or build_path('tree'))
    if not out.parent.exists():
        out.parent.mkdir()
    if out.exists():
        rmtree(out)
    out.mkdir()
    old_tree = {l.id: l for l in walk_tree(tree)} if tree else {}

    languages = {}
    for lang in read_lff('language'):
        languages[lang.id] = lang
        lang2tree(lang, lang.lineage, out, old_tree)

    for lang in read_lff('dialect'):
        if not lang.lineage or lang.lineage[0][1] not in languages:
            raise ValueError('unattached dialect')

        lang2tree(
            lang, languages[lang.lineage[0][1]].lineage + lang.lineage, out, old_tree)

    if not test:
        rmtree(TREE, ignore_errors=True)
        copytree(out, TREE)


def tree2lff(tree=TREE):
    languoids = dict(dialect=defaultdict(list), language=defaultdict(list))
    nodes = {}

    for l in walk_tree(tree=tree, nodes=nodes):
        if l.level in languoids:
            languoids[l.level][l.lff_group()].append(l.lff_language())

    for level, languages in languoids.items():
        with build_path('%sff.txt' % level[0]).open('w', encoding='utf8') as fp:
            fp.write('# -*- coding: utf-8 -*-\n')
            for path in sorted(languages):
                fp.write(path + '\n')
                for l in sorted(languages[path]):
                    fp.write(l + '\n')



# coding: utf8
"""
Main command line interface of the pyglottolog package.

Like programs such as git, this cli splits its functionality into sub-commands
(see e.g. https://docs.python.org/2/library/argparse.html#sub-commands).
The rationale behind this is that while a lot of different tasks may be triggered using
this cli, most of them require common configuration.

The basic invocation looks like

    glottolog [OPTIONS] <command> [args]

"""
from __future__ import unicode_literals
import sys
import argparse

from pyglottolog.monster import main as compile_monster
from pyglottolog.languoids import make_index
from pyglottolog import lff


def monster(args):
    """Compile the monster bibfile from the BibTeX files listed in references/BIBFILES.ini

    glottolog monster
    """
    compile_monster()


def index(args):
    """Create an index page listing and linking to all languoids of a specified level.

    glottolog index (family|language|dialect|all)
    """
    for level in ['family', 'language', 'dialect']:
        if args.args[0] in [level, 'all']:
            make_index(level)


def tree2lff(args):
    """Create lff.txt and dff.txt from the current languoid tree.

    glottolog tree2lff
    """
    lff.tree2lff()


def lff2tree(args):
    """Recreate tree from lff.txt and dff.txt

    glottolog lff2tree [test]
    """
    lff.lff2tree(test=args.args and args.args[0] == 'test')
    if args.args and args.args[0] == 'test':
        print("""
You can run

    diff -rbB build/tree/ languoids/tree/

to inspect the changes in the directory tree.
""")
    else:
        print("""
Run

    git status

to inspect changes in the directory tree.

- To discard changes run

    git checkout languoids/tree

- To commit and push changes, run

    git add languoids/tree/...

  for any newly created nodes listed under

# Untracked files:
#   (use "git add <file>..." to include in what will be committed)
#
#	languoids/tree/...

  followed by

    git commit -a -m"reason for change of classification"
    git push origin
""")


COMMANDS = {f.__name__: f for f in [monster, index, tree2lff, lff2tree]}


def main():
    parser = argparse.ArgumentParser(
        description="""Main command line interface of the pyglottolog package.""",
        epilog="Use '%(prog)s help <cmd>' to get help about individual commands.")
    parser.add_argument("--verbosity", help="increase output verbosity")
    parser.add_argument('command', help='|'.join(COMMANDS))
    parser.add_argument('args', nargs=argparse.REMAINDER)

    args = parser.parse_args()
    if args.command == 'help':
        # As help text for individual commands we simply re-use the docstrings of the
        # callables registered for the command:
        print(COMMANDS[args.args[0]].__doc__)
        sys.exit(0)

    COMMANDS[args.command](args)
    sys.exit(0)



# coding: utf8
from __future__ import unicode_literals
import re
from itertools import takewhile
from collections import defaultdict

from clldutils.misc import slug
from clldutils.path import Path, walk
from clldutils.inifile import INI

from pyglottolog.util import languoids_path, parse_conjunctions


TREE = languoids_path('tree')
ID_REGEX = '([a-z0-9]{4}[0-9]{4}|NOCODE(_[A-Za-z0-9\-]+)?)'


class Languoid(object):
    section_core = 'core'
    id_pattern = re.compile(ID_REGEX + '$')

    def __init__(self, cfg, lineage=None, directory=None):
        """

        :param cfg:
        :param lineage: list of ancestors, given as (id, name) pairs.
        """
        lineage = lineage or []
        assert all([self.id_pattern.match(id) for name, id, level in lineage])
        self.lineage = lineage
        self.cfg = cfg
        self.dir = directory or TREE.joinpath(*[id for name, id, _ in self.lineage])

    @classmethod
    def from_dir(cls, directory, **kw):
        for p in directory.iterdir():
            if p.is_file():
                assert p.suffix == '.ini'
                return cls.from_ini(p, **kw)

    @classmethod
    def from_ini(cls, ini, nodes={}):
        if not isinstance(ini, Path):
            ini = Path(ini)

        directory = ini.parent
        cfg = INI(interpolation=None)
        cfg.read(ini.as_posix(), encoding='utf8')

        lineage = []
        for parent in directory.parents:
            id_ = parent.name
            assert id_ != directory.name
            if not cls.id_pattern.match(id_):
                # we ignore leading non-languoid-dir path components.
                break

            if id_ not in nodes:
                l = Languoid.from_dir(parent, nodes=nodes)
                nodes[id_] = (l.name, l.id, l.level)
            lineage.append(nodes[id_])

        res = cls(cfg, list(reversed(lineage)), directory=directory)
        nodes[res.id] = (res.name, res.id, res.level)
        return res

    @classmethod
    def from_name_id_level(cls, name, id, level):
        cfg = INI(interpolation=None)
        cfg.read_dict(dict(core=dict(name=name, glottocode=id, level=level)))
        return cls(cfg, [])

    @classmethod
    def from_lff(cls, path, name_and_codes, level):
        lname, codes = name_and_codes.split('[', 1)
        lname = lname.strip()
        glottocode, isocode = codes[:-1].split('][')

        lineage = []
        if path:
            for i, comp in enumerate(path.split('], ')):
                if comp.endswith(']'):
                    comp = comp[:-1]
                name, id_ = comp.split(' [', 1)
                _level = 'family'
                if level == 'dialect':
                    _level = 'language' if i == 0 else 'dialect'
                lineage.append((name, id_, _level))

        cfg = INI(interpolation=None)
        cfg.read_dict(dict(core=dict(name=lname, glottocode=glottocode, level=level)))
        res = cls(cfg, lineage)
        if isocode:
            res.iso = isocode
        return res

    @property
    def children(self):
        return [Languoid.from_dir(d) for d in self.dir.iterdir() if d.is_dir()]

    @property
    def ancestors(self):
        res = []
        for parent in self.dir.parents:
            id_ = parent.name
            if self.id_pattern.match(id_):
                res.append(Languoid.from_dir(parent))
            else:
                # we ignore leading non-languoid-dir path components.
                break
        return list(reversed(res))

    @property
    def parent(self):
        ancestors = self.ancestors
        return ancestors[-1] if ancestors else None

    @property
    def family(self):
        ancestors = self.ancestors
        return ancestors[0] if ancestors else None

    def _set(self, key, value):
        self.cfg.set(self.section_core, key, value)

    def _get(self, key, type_=None):
        res = self.cfg.get(self.section_core, key, fallback=None)
        if type_ and res:
            return type_(res)
        return res

    @property
    def macroareas(self):
        return self.cfg.getlist(self.section_core, 'macroareas')

    @macroareas.setter
    def macroareas(self, value):
        assert isinstance(value, (list, tuple))
        self._set('macroareas', value)

    @property
    def name(self):
        return self._get('name')

    @name.setter
    def name(self, value):
        self._set('name', value)

    @property
    def id(self):
        return self._get('glottocode')

    @id.setter
    def id(self, value):
        self._set('glottocode', value)

    @property
    def glottocode(self):
        return self._get('glottocode')

    @glottocode.setter
    def glottocode(self, value):
        self._set('glottocode', value)

    @property
    def latitude(self):
        return self._get('latitude', float)

    @latitude.setter
    def latitude(self, value):
        self._set('latitude', value)

    @property
    def longitude(self):
        return self._get('longitude', float)

    @longitude.setter
    def longitude(self, value):
        self._set('longitude', value)

    @property
    def hid(self):
        return self._get('hid')

    @property
    def level(self):
        return self._get('level')

    @level.setter
    def level(self, value):
        self._set('level', value)

    @property
    def iso(self):
        return self._get('iso639-3')

    @iso.setter
    def iso(self, value):
        self._set('iso639-3', value)

    @property
    def iso_code(self):
        return self._get('iso639-3')

    @iso_code.setter
    def iso_code(self, value):
        self._set('iso639-3', value)

    @property
    def classification_status(self):
        return self._get('classification_status')

    @classification_status.setter
    def classification_status(self, value):
        self._set('classification_status', value)

    def fname(self, suffix=''):
        return '%s%s' % (self.id, suffix)

    def write_info(self, outdir):
        if not isinstance(outdir, Path):
            outdir = Path(outdir)
        self.cfg.write(outdir.joinpath(self.fname('.ini')))

    def lff_group(self):
        if self.level == 'dialect':
            lineage = reversed(
                list(takewhile(lambda x: x[2] != 'family', reversed(self.lineage))))
        else:
            lineage = self.lineage
        if not self.lineage:
            return '%s [-isolate-]' % self.name
        res = ', '.join('%s [%s]' % (name, id) for name, id, level in lineage)
        return res or 'ERROR [-unclassified-]'

    def lff_language(self):
        res = '    %s [%s][%s]' % (self.name, self.id, self.iso or '')
        if self.classification_status:
            res = '%s %s' % (res, self.classification_status)
        return res


def walk_tree(tree=TREE, **kw):
    for fname in walk(tree, mode='files', followlinks=True):
        if fname.suffix == '.ini':
            yield Languoid.from_ini(fname, **kw)


def make_index(level):
    fname = dict(language='languages', family='families', dialect='dialects')[level]
    links = defaultdict(dict)
    for lang in walk_tree():
        if lang.level == level:
            label = '{0.name} [{0.id}]'.format(lang)
            if lang.iso:
                label += '[%s]' % lang.iso
            links[slug(lang.name)[0]][label] = \
                lang.dir.joinpath(lang.fname('.ini')).relative_to(languoids_path())

    with languoids_path(fname + '.md').open('w', encoding='utf8') as fp:
        fp.write('## %s\n\n' % fname.capitalize())
        fp.write(' '.join(
            '[-%s-](%s_%s.md)' % (i.upper(), fname, i) for i in sorted(links.keys())))
        fp.write('\n')

    for i, langs in links.items():
        with languoids_path('%s_%s.md' % (fname, i)).open('w', encoding='utf8') as fp:
            for label in sorted(langs.keys()):
                fp.write('- [%s](%s)\n' % (label, langs[label]))


#
# The following two functions are necessary to make the compilation of the monster bib
# compatible with the new way of storing languoid data.
#
def macro_area_from_hid(tree=TREE):
    res = {}
    for lang in walk_tree(tree):
        if lang.hid:
            macroareas = lang.cfg.getlist('core', 'macroareas')
            res[lang.hid] = macroareas[0] if macroareas else ''
    return res


def load_triggers(tree=TREE, type_='lgcode'):
    res = {}
    for lang in walk_tree(tree):
        if lang.hid and lang.cfg.has_option('triggers', type_):
            triggers = lang.cfg.getlist('triggers', type_)
            if not triggers:
                continue
            res[(type_, '%s [%s]' % (lang.name, lang.hid))] = [
                parse_conjunctions(t) for t in triggers]
    return res



# _bibtex_undiacritic.py - remove latex/unicode diacritics for string comparison

import re

from unidecode import unidecode

__all__ = ['undiacritic']


class Replace(object):
    """Multiple search-replace with mutually exclusive regexes."""

    _rules = [  # ordered counterbleeding
        (r'\\AA(?:\{\})?', 'A'),
        (r'\\AE(?:\{\})?', 'Ae'),
        (r'\\aa(?:\{\})?', 'a'),
        (r'\\ae(?:\{\})?', 'e'),
        (r'\\oslash(?:\{\})?', 'o'),
        (r'\\Oslash(?:\{\})?', 'O'),
        (r'\\OE(?:\{\})?', 'OE'),
        (r'\\oe(?:\{\})?', 'oe'),
        (r'\\O(?:\{\})?', 'O'),
        (r'\\o(?:\{\})?', 'o'),
        (r'\\L(?:\{\})?', 'L'),
        (r'\\l(?:\{\})?', 'l'),
        (r'\\i(?:\{\})?', 'i'),
        (r'\\j(?:\{\})?', 'j'),
        (r'\\NG(?:\{\})?', 'NG'),
        (r'\\ng(?:\{\})?', 'ng'),
        (r'\\texteng(?:\{\})?', 'ng'),
        (r'\\ss(?:\{\})?', 'ss'),
        (r'\\textbari(?:\{\})?', 'i'),
        (r'\\textbaru(?:\{\})?', 'u'),
        (r'\\textbarI(?:\{\})?', 'I'),
        (r'\\textbarU(?:\{\})?', 'U'),
        (r'\\texthtd(?:\{\})?', 'd'),
        (r'\\texthtb(?:\{\})?', 'b'),
        (r'\\textopeno(?:\{\})?', 'o'),
        (r'\\textepsilon(?:\{\})?', 'e'),
        (r'\\textschwa(?:\{\})?', 'e'),
        (r'\\textrhooktopd(?:\{\})?', 'd'),
        (r'\\textthorn(?:\{\})?', 'th'),
    ]

    def __init__(self, pairs=_rules):
        self._old, self._new = zip(*pairs)
        self._pattern = re.compile('|'.join('(%s)' % o for o in self._old))

    def __call__(self, s):
        return self._pattern.sub(self._repl, s)

    def _repl(self, match):
        return self._new[match.lastindex - 1]


REPLACE = Replace()
COMMAND1 = re.compile(r'\\text[a-z]+\{([^}]*)\}')
COMMAND2 = re.compile(r'\\text[a-z]+')
ACCENT = re.compile(r'''\\[`'^"H~ckl=b.druvt](\{[a-zA-Z]\}|[a-zA-Z])''')
DROP = re.compile(r'\\[^\s{}]+\{|\\.|[{}]')


def undiacritic(txt):
    if isinstance(txt, unicode):
        txt = unidecode(txt)
    txt = REPLACE(txt)
    txt = COMMAND1.sub(r'\1', txt)
    txt = COMMAND2.sub('', txt)
    txt = ACCENT.sub(r'\1', txt)
    return DROP.sub('', txt)



# _compare_names.py - compare regex-based with pybtex name parsing

import _bibfiles, _bibtex, _libmonster


def names1(s):
    return [(debr(n.get('firstname', '')), debr(n.get('lastname', '')), debr(n.get('jr', '')))
        for n in _libmonster.pauthor(s)]


def names2(s):
    return [(debr(first), debr(' '.join(n for n in (prelast, last) if n)), debr(lineage))
        for prelast, last, first, lineage in _bibtex.names(s)]


def debr(s):
    if s.startswith('{') and s.endswith('}'):
        return s[1:-1]
    return s


counts = {}

for b in _bibfiles.Collection():
    print b.filename.center(80, '#')
    count = 0
    for bibkey, (entrytype, fields) in b.iterentries():
        for role in ('author', 'editor'):
            names = fields.get(role, '')
            n1, n2 = names1(names), names2(names)
            if n1 != n2:
                count += 1
                print b.filename, bibkey, role
                print repr(names)
                print n1
                print n2
                print
    counts[b.filename] = count

print '\n'.join('%d\t%s' % (n, f) for f, n in sorted(counts.iteritems()))



# _libmonster.py - mixed support library

# TODO: consider replacing pauthor in keyid with _bibtex.names
# TODO: enusure \emph is dropped from titles in keyid calculation

import re
import csv
from collections import namedtuple
from heapq import nsmallest

from pyglottolog import languoids
from pyglottolog.util import references_path, parse_conjunctions, read_ini
from pyglottolog._bibtex_undiacritic import undiacritic

__all__ = [
    'add_inlg_e',
    'keyid',
    'wrds', 'setd', 'setd3', 'indextrigs',
    'lstat', 'lstat_witness', 
    'hhtype_to_n', 'expl_to_hhtype', 'lgcode',
    'read_csv_dict', 'write_csv_rows', 'load_triggers',
    'pitems',
]

HHTYPE = references_path('alt4hhtype.ini')


def read_csv_dict(filename):
    return {row[0]: row for row in csv_iterrows(filename)}


def csv_iterrows(filename, fieldnames=None, dialect='excel'):
    with open(filename) as fd:
        reader = csv.reader(fd, dialect=dialect)
        if fieldnames is None:
            fieldnames = next(reader)
        make_row = namedtuple('Row', fieldnames)._make
        for row in reader:
            yield make_row(row)


def write_csv_rows(rows, filename, fieldnames=None, encoding='utf-8', dialect='excel'):
    with open(filename, 'wb') as fd:
        writer = csv.writer(fd, dialect=dialect)
        if fieldnames:
            writer.writerow([unicode(f).encode(encoding) for f in fieldnames])
        writer.writerows([[unicode(c).encode(encoding) for c in r] for r in rows])


def load_triggers(filename):
    p = read_ini(filename)
    result = {}
    for s in p.sections():
        cls, lab = s.split(', ', 1)
        triggers = p.get(s, 'triggers').strip().splitlines()
        if not triggers:  # hhtype, unknown
            continue
        result[(cls, lab)] = [parse_conjunctions(t) for t in triggers]
    return result


def load_hhtypes(filename=HHTYPE):
    p = read_ini(filename)
    result = {}
    for s in p.sections():
        _, _, expl = s.partition(', ')
        hht = p.get(s, 'id')
        rank = p.getint(s, 'rank')
        abbv = p.get(s, 'abbv')
        bibabbv = p.get(s, 'bibabbv')
        result[hht] = (rank, expl, abbv, bibabbv)
    return result


def setd3(ds, k1, k2, k3, v=None):
    if ds.has_key(k1):
        if ds[k1].has_key(k2):
            ds[k1][k2][k3] = v
        else:
            ds[k1][k2] = {k3: v}
    else:
        ds[k1] = {k2: {k3: v}}
    return


def setd(ds, k1, k2, v=None):
    if ds.has_key(k1):
        ds[k1][k2] = v
    else:
        ds[k1] = {k2: v}
    return


def opv(d, func):
    return {i: func(v) for i, v in d.iteritems()}


def grp2fd(l):
    r = {}
    for (a, b) in l:
        if r.has_key(a):
            r[a][b] = r[a].get(b, 0) + 1
        else:
            r[a] = {b: 1}
    return r


def grp2(l):
    r = {}
    for (a, b) in l:
        setd(r, a, b)
    return opv(r, lambda x: x.keys())


reauthor = [re.compile(pattern) for pattern in [
    "(?P<lastname>[^,]+),\s((?P<jr>[JS]r\.|[I]+),\s)?(?P<firstname>[^,]+)$",
    "(?P<firstname>[^{][\S]+(\s[A-Z][\S]+)*)\s(?P<lastname>([a-z]+\s)*[A-Z\\\\][\S]+)(?P<jr>,\s[JS]r\.|[I]+)?$",
    "(?P<firstname>\\{[\S]+\\}[\S]+(\s[A-Z][\S]+)*)\s(?P<lastname>([a-z]+\s)*[A-Z\\\\][\S]+)(?P<jr>,\s[JS]r\.|[I]+)?$",
    "(?P<firstname>[\s\S]+?)\s\{(?P<lastname>[\s\S]+)\}(?P<jr>,\s[JS]r\.|[I]+)?$",
    "\{(?P<firstname>[\s\S]+)\}\s(?P<lastname>[\s\S]+?)(?P<jr>,\s[JS]r\.|[I]+)?$",
    "(?P<lastname>[A-Z][\S]+)$",
    "\{(?P<lastname>[\s\S]+)\}$",
    "(?P<lastname>[aA]nonymous)$",
    "(?P<lastname>\?)$",
    "(?P<lastname>[\s\S]+)$",
]]

def psingleauthor(n, vonlastname=True):
    for pattern in reauthor:
        o = pattern.match(n)
        if o:
            if vonlastname:
                return lastvon(o.groupdict())
            return o.groupdict()
    if n:
        print "Couldn't parse name:", n
    return None


def pauthor(s):
    pas = [psingleauthor(a) for a in s.split(' and ')]
    if [a for a in pas if not a]:
        if s:
            print s
    return [a for a in pas if a]


#"Adam, A., W.B. Wood, C.P. Symons, I.G. Ord & J. Smith"
#"Karen Adams, Linda Lauck, J. Miedema, F.I. Welling, W.A.L. Stokhof, Don A.L. Flassy, Hiroko Oguri, Kenneth Collier, Kenneth Gregerson, Thomas R. Phinnemore, David Scorza, John Davies, Bernard Comrie & Stan Abbott"


relu = re.compile("\s+|(d\')(?=[A-Z])")
recapstart = re.compile("\[?[A-Z]")

def lowerupper(s):
    parts = [x for x in relu.split(s) if x]
    lower = []
    upper = []
    for (i, x) in enumerate(parts):
        if not recapstart.match(undiacritic(x)):
            lower.append(x)
        else:
            upper = parts[i:]
            break
    return (lower, upper)


def lastvon(author):
    if not author.has_key('firstname'):
        return author
    r = {}
    (lower, upper) = lowerupper(author['firstname'])
    r['lastname'] = (' '.join(lower).strip() + ' ' + author['lastname']).strip()
    r['firstname'] = ' '.join(upper)
    if author.has_key('jr') and author['jr']:
        r['jr'] = author['jr']

    return r


def lastnamekey(s):
    (_, upper) = lowerupper(s)
    if not upper:
        return ''
    return max(upper)


def rangecomplete(incomplete, complete):
    if len(complete) > len(incomplete):
        return complete[:len(complete)-len(incomplete)] + incomplete
    return incomplete


rebracketyear = re.compile("\[([\d\,\-\/]+)\]")
reyl = re.compile("[\,\-\/\s\[\]]+")

def pyear(s):
    if rebracketyear.search(s):
        s = rebracketyear.search(s).group(1)
    my = [x for x in reyl.split(s) if x.strip()]
    if len(my) == 0:
        return "[nd]"
    if len(my) != 1:
        return my[0] + "-" + rangecomplete(my[-1], my[0])
    return my[-1]


refields = re.compile('\s*(?P<field>[a-zA-Z\_]+)\s*=\s*[{"](?P<data>.*)[}"],\r?\n')
refieldsnum = re.compile('\s*(?P<field>[a-zA-Z\_]+)\s*=\s*(?P<data>\d+),\r?\n')
refieldsacronym = re.compile('\s*(?P<field>[a-zA-Z\_]+)\s*=\s*(?P<data>[A-Za-z]+),\r?\n')
#refieldslast = re.compile('\s*(?P<field>[a-zA-Z\_]+)\s*=\s*[{"]*(?P<data>.+?)[}"]*\r?\n}')
refieldslast = re.compile('\s*(?P<field>[a-zA-Z\_]+)\s*=\s*[\{\"]?(?P<data>[^\r\n]+?)[\}\"]?(?<!\,)(?:$|\r?\n)')
retypekey = re.compile("@(?P<type>[a-zA-Z]+){(?P<key>[^,\s]*)[,\r\n]")
reitem = re.compile("@[a-zA-Z]+{[^@]+}")

trf = '@Book{g:Fourie:Mbalanhu,\n  author =   {David J. Fourie},\n  title =    {Mbalanhu},\n  publisher =    LINCOM,\n  series =       LWM,\n  volume =       03,\n  year = 1993\n}'

def pitems(txt):
    for m in reitem.finditer(txt):
        item = m.group()
        o = retypekey.search(item)
        if o is None:
            continue
        key = o.group("key")
        typ = o.group("type")
        fields = refields.findall(item) + refieldsacronym.findall(item) + refieldsnum.findall(item) + refieldslast.findall(item)
        fieldslower = ((x.lower(), y) for x, y in fields)
        yield key, typ.lower(), dict(fieldslower)


#	Author = ac # { and Howard Coate},
#	Author = ad,


bibord = {k: i for i, k in enumerate([
    'author',
    'editor',
    'title',
    'booktitle',
    'journal',
    'school',
    'publisher',
    'address',
    'series',
    'volume',
    'number',
    'pages',
    'year',
    'issn',
    'url',
])}

def bibord_iteritems(fields, sortkey=lambda f, inf=float('inf'): (bibord.get(f, inf), f)):
    for f in sorted(fields, key=sortkey):
        yield f, fields[f]


resplittit = re.compile("[\(\)\[\]\:\,\.\s\-\?\!\;\/\~\=]+")

def wrds(txt):
    txt = undiacritic(txt.lower())
    txt = txt.replace("'", "").replace('"', "")
    return [x for x in resplittit.split(txt) if x]


def renfn(e, ups):
    for (k, field, newvalue) in ups:
        (typ, fields) = e[k]
        #fields['mpifn'] = fields['fn']
        fields[field] = newvalue
        e[k] = (typ, fields)
    return e


def add_inlg_e(e):
    inlg = languoids.load_triggers(type_='inlg')
    # FIXME: does not honor 'NOT' for now
    dh = {word: label  for (cls, label), triggers in inlg.iteritems()
        for t in triggers for flag, word in t}  
    ts = [(k, wrds(fields['title']) + wrds(fields.get('booktitle', ''))) for (k, (typ, fields)) in e.iteritems() if fields.has_key('title') and not fields.has_key('inlg')]
    print len(ts), "without", 'inlg'
    ann = [(k, set(dh[w] for w in tit if dh.has_key(w))) for (k, tit) in ts]
    unique = [(k, lgs.pop()) for (k, lgs) in ann if len(lgs) == 1]
    print len(unique), "cases of unique hits"
    fnups = [(k, 'inlg', v) for (k, v) in unique]
    t2 = renfn(e, fnups)
    #print len(unique), "updates"

    newtrain = grp2fd([(lgcodestr(fields['inlg'])[0], w) for (k, (typ, fields)) in t2.iteritems() if fields.has_key('title') and fields.has_key('inlg') if len(lgcodestr(fields['inlg'])) == 1 for w in wrds(fields['title'])])
    #newtrain = grp2fd([(cname(lgc), w) for (lgcs, w) in alc if len(lgcs) == 1 for lgc in lgcs])
    for (lg, wf) in sorted(newtrain.iteritems(), key=lambda x: len(x[1])):
        cm = [(1+f, float(1-f+sum(owf.get(w, 0) for owf in newtrain.itervalues())), w) for (w, f) in wf.iteritems() if f > 9]
        cms = [(f/fn, f, fn, w) for (f, fn, w) in cm]
        cms.sort(reverse=True)
        ##print lg, cms[:10]
        ##print ("h['%s'] = " % lg) + str([x[3] for x in cms[:10]])
    return t2


rerpgs = re.compile("([xivmcl]+)\-?([xivmcl]*)")
repgs = re.compile("([\d]+)\-?([\d]*)")

def pagecount(pgstr):
    rpgs = rerpgs.findall(pgstr)
    pgs = repgs.findall(pgstr)
    rsump = sum([romanint(b) - romanint(a) + 1 for (a, b) in rpgs if b] + [romanint(a) for (a, b) in rpgs if not b])
    sump = sum([int(rangecomplete(b, a)) - int(a) + 1 for (a, b) in pgs if b] + [int(a) for (a, b) in pgs if not b])
    if rsump != 0 and sump != 0:
        return "%s+%s" % (rsump, sump)
    if rsump == 0 and sump == 0:
        return ''
    return str(rsump + sump)


def introman(i):
    z = {'m': 1000, 'd': 500, 'c': 100, 'l': 50, 'x': 10, 'v': 5, 'i': 1}
    iz = dict((v, k) for (k, v) in z.iteritems())
    x = ""
    for (v, c) in sorted(iz.items(), reverse=True):
        (q, r) = divmod(i, v)
        if q == 4 and c != 'm':
            x = x + c + iz[5*v]
        else:
            x = x + ''.join(c for i in range(q))
        i = r
    return x


def romanint(r):
    z = {'m': 1000, 'd': 500, 'c': 100, 'l': 50, 'x': 10, 'v': 5, 'i': 1}
    i = 0
    prev = 10000
    for c in r:
        zc = z[c]
        if zc > prev:
            i = i - 2*prev + zc
        else:
            i = i + zc
        prev = zc
    return i


rerom = re.compile("(\d+)")

def roman(x):
    return rerom.sub(lambda o: introman(int(o.group(1))), x).upper()


rewrdtok = re.compile("[a-zA-Z].+")
reokkey = re.compile("[^a-z\d\-\_\[\]]")

def keyid(fields, fd={}, ti=2, infinity=float('inf')):
    if not fields.has_key('author'):
        if not fields.has_key('editor'):
            values = ''.join(v for f, v in bibord_iteritems(fields)
                if f != 'glottolog_ref_id')
            return '__missingcontrib__' + reokkey.sub('_', values.lower())
        else:
            astring = fields['editor']
    else:
        astring = fields['author']

    authors = pauthor(astring)
    if len(authors) != len(astring.split(' and ')):
        print "Unparsed author in", authors
        print "   ", astring, astring.split(' and ')
        print fields['title']

    ak = [undiacritic(x) for x in sorted(lastnamekey(a['lastname']) for a in authors)]
    yk = pyear(fields.get('year', '[nd]'))[:4]
    tks = wrds(fields.get("title", "no.title")) #takeuntil :
    # select the (leftmost) two least frequent words from the title
    types = uniqued(w for w in tks if rewrdtok.match(w))
    tk = nsmallest(ti, types, key=lambda w: fd.get(w, infinity))
    # put them back into the title order (i.e. 'spam eggs' != 'eggs spam')
    order = {w: i for i, w in enumerate(types)}
    tk.sort(key=lambda w: order[w])
    if fields.has_key('volume') and not fields.has_key('journal') and not fields.has_key('booktitle') and not fields.has_key('series'):
        vk = roman(fields['volume'])
    else:
        vk = ''

    if fields.has_key('extra_hash'):
        yk = yk + fields['extra_hash']

    key = '-'.join(ak) + "_" + '-'.join(tk) + vk + yk
    return reokkey.sub("", key.lower())


def uniqued(items):
    seen = set()
    return [i for i in items if i not in seen and not seen.add(i)]


reisobrack = re.compile("\[([a-z][a-z][a-z]|NOCODE\_[A-Z][^\s\]]+)\]")
recomma = re.compile("[\,\/]\s?")
reiso = re.compile("[a-z][a-z][a-z]$|NOCODE\_[A-Z][^\s\]]+$")

def lgcode((typ, fields)):
    if not fields.has_key('lgcode'):
        return []
    return lgcodestr(fields['lgcode'])


def lgcodestr(lgcstr):
    lgs = reisobrack.findall(lgcstr)
    if lgs:
        return lgs

    parts = [p.strip() for p in recomma.split(lgcstr)]
    codes = [p for p in parts if reiso.match(p)]
    if len(codes) == len(parts):
        return codes
    return []


rekillparen = re.compile(" \([^\)]*\)")
respcomsemic = re.compile("[;,]\s?")

def hhtypestr(s):
    return respcomsemic.split(rekillparen.sub("", s))


def indextrigs(ts):
    return grp2([(tuple(sorted(disj)), clslab) for (clslab, t) in ts.iteritems() for disj in t])


def sd(es):
    #most signficant piece of descriptive material
    #hhtype, pages, year
    mi = [(k, (hhtypestr(fields.get('hhtype', 'unknown')), fields.get('pages', ''), fields.get('year', ''))) for (k, (typ, fields)) in es.iteritems()]
    d = accd(mi)
    ordd = [sorted(((p, y, k, t) for (k, (p, y)) in d[t].iteritems()), reverse=True) for t in hhtyperank if d.has_key(t)]
    return ordd


def pcy(pagecountstr):
    #print pagecountstr
    if not pagecountstr:
        return 0
    return eval(pagecountstr) #int(takeafter(pagecountstr, "+"))


def accd(mi):
    r = {}
    for (k, (hhts, pgs, year)) in mi:
        #print k
        pci = pcy(pagecount(pgs))
        for t in hhts:
            setd(r, t, k, (pci/float(len(hhts)), year))
    return r


def byid(es, idf=lgcode, unsorted=False):
    def tftoids(tf):
        z = idf(tf)
        if unsorted and not z:
            return ['!Unsorted']
        return z
    return grp2([(cfn, k) for (k, tf) in es.iteritems() for cfn in tftoids(tf)])


hhtypes = load_hhtypes()
hhtyperank = [hht for (n, expl, abbv, bibabbv, hht) in sorted((info + (hht,) for (hht, info) in hhtypes.iteritems()), reverse=True)]
hhtype_to_n = dict((x, len(hhtyperank)-i) for (i, x) in enumerate(hhtyperank))
expl_to_hhtype = dict((expl, hht) for (hht, (n, expl, abbv, bibabbv)) in hhtypes.iteritems())


def sdlgs(e, unsorted=False):
    eindex = byid(e, unsorted=unsorted)
    fes = opv(eindex, lambda ks: dict((k, e[k]) for k in ks))
    fsd = opv(fes, sd)
    return (fsd, fes)


def lstat(e, unsorted=False):
    (lsd, lse) = sdlgs(e, unsorted=unsorted)
    return opv(lsd, lambda xs: (xs + [[[None]]])[0][0][-1])


def lstat_witness(e, unsorted=False):
    def statwit(xs):
        if len(xs) == 0:
            return (None, [])
        [(typ, ks)] = grp2([(t, k) for [p, y, k, t] in xs[0]]).items()
        return (typ, ks)
    (lsd, lse) = sdlgs(e, unsorted=unsorted)
    return opv(lsd, statwit)



from setuptools import setup, find_packages


requires = [
    # list required third-party packages here
    'clldutils>=0.7',
    'pybtex',
    'latexcodec',
    'unidecode',
]

setup(
    name='pyglottolog',
    version='0.1',
    description='python package for glottolog data curation',
    long_description='',
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
    ],
    author='',
    author_email='',
    url='',
    keywords='data linguistics',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    entry_points={
        'console_scripts': ['glottolog=pyglottolog.cli:main'],
    },
    tests_require=[],
    test_suite="pyglottolog")



# _bibtex_escaping.py - latex escapes and unicode bracket escapes

"""TODO: replace in source files OR add de-/encoding rule

\textsubdot{
\textsubbar{
\textsuperscript{ c
\~e
\textsubline{
\texthtd
\textvertline
\textesh
\textbaru
\textbari
\textsubring
\textgamma
\textupsilon


\texthooktop{
\textsubwedge{
\textemdash
\textendash
\textthreequarters
\texthooktopd
\textrevglotstop
\;
\:

---, --, ``'', {}, \textdot, \textsubdot,  \v{j} -> \v\j

see also:
  _bibtex_undiacritic.py rules
  http://github.com/clld/clld/blob/master/clld/lib/bibtex.py
  http://github.com/clld/clld/blob/master/clld/lib/latex.py
  http://github.com/mcmtroffaes/latexcodec/blob/develop/latexcodec/codec.py#L97

FIXME: tex-escape/character entity mix in anla.bib titlealt

\textpolhook{ -> \k{}?
\&
"""

import re
import codecs

import latexcodec

__all__ = ['u_escape', 'u_unescape', 'latex_to_utf8']

LATEX_TABLE = {
    u'\N{LATIN SMALL LETTER ENG}': br'\ng',
    u'\N{LATIN CAPITAL LETTER ENG}': br'\NG',
    u'\N{LATIN SMALL LETTER OPEN O}': br'\textopeno',
    u'\N{LATIN SMALL LETTER THORN}': br'\textthorn',
    u'\N{LATIN SMALL LETTER GLOTTAL STOP}': br'\textglotstop',
    u'\N{MODIFIER LETTER GLOTTAL STOP}': br'\textraiseglotstop',
    u'\N{LATIN SMALL LETTER I WITH STROKE}': br'\textbari',
}

_LC_TABLES = (latexcodec.codec._LATEX_UNICODE_TABLE,)

assert u'\xe4'.encode('latex') == r'\"a'
assert r'\"a'.decode('latex') == u'\xe4'

for unicode_text, latex_text in LATEX_TABLE.iteritems():
    for table in _LC_TABLES:
        table.register(unicode_text, latex_text)


def u_escape(s):
    def iterchunks(s):
        for c in s:
            o = ord(c)
            if o <= 127:
                yield c
            else:
                yield r'?[\u%d]' % o

    return ''.join(iterchunks(s))


def u_unescape(s, pattern=re.compile(r'\?\[\\u(\d{3,5})\]')):
    def iterchunks(s, matches):
        pos = 0
        for m in matches:
            start, end = m.span()
            yield s[pos:start]
            yield unichr(int(m.group(1)))
            pos = end
        yield s[pos:]

    return ''.join(iterchunks(s, pattern.finditer(s)))


class U_escapeCodec(codecs.Codec):

    def encode(self, input, erors='strict'):
        return u_escape(input.encode('ascii')), len(input)

    def decode(self, input, erors='strict'):
        return u_unescape(input.decode('ascii')), len(input)


class U_escapeStreamWriter(U_escapeCodec, codecs.StreamWriter):
    pass


class U_escapeStreamReader(U_escapeCodec, codecs.StreamReader):
    pass


def _find_u_escape(encoding):
    if encoding == 'ascii+u_escape':
        return codecs.CodecInfo(name='ascii+u_escape',
            encode=U_escapeCodec().encode, decode=U_escapeCodec().decode,
            streamwriter=U_escapeStreamWriter, streamreader=U_escapeStreamReader)


codecs.register(_find_u_escape)


# legacy

def latex_to_utf8(s, verbose=True, debracket=re.compile("\{(.)\}")):
    us = s.decode("latex")
    us = debracket.sub("\\1", us)
    if verbose:
        remaininglatex(us)
    return us


platexspc = [re.compile(pattern) for pattern in [
    r'''\\(?P<typ>[^\%'\`\^\~\=\_\"\s\{]+)\{(?P<ch>[a-zA-Z]?)\}''',
    r'''\\(?P<typ>['\`\^\~\=\_\"]+?)\{(?P<ch>[a-zA-Z])\}''',
    r'\\(?P<typ>[^a-zA-Z\s\%\_])(?P<ch>[a-zA-Z])',
    r'\\(?P<typ>[^a-zA-Z\s\%\{\_]+)(?P<ch>[a-zA-Z])',
    r'\\(?P<typ>[^\{\%\_]+)\{(?P<ch>[^\}]+)\}',
    r'\\(?P<typ>[^\{\_\\\s\%]+)(?P<ch>\s)',
]]


def remaininglatex(txt):
    for pattern in platexspc:
        o = pattern.findall(txt)
        if o:
            print o[:100] #txt[o.start()-10:o.start()+10], o.groups()
